import logging
import random
import selenium
import time

LOGGER = logging.getLogger('wabot')

class PageObject:
    """
    Wrapper around page element that provides an object orientated interface
    to elements. Can be sublcassed to integrate more complicated instruction.
    """
    def __init__(self, page, accessors=None, name=None):
        self.page = page
        self.driver = page.driver
        self.accessors = accessors
        self.name = name
        # self.el = None

    def __getattr__(self, name):
        # if not self.el:
        #     raise AttributeError
        try:
            return getattr(self.el, name) 
        except AttributeError as ex:
            raise

    def selenium_element(self, accessors=None):
        """
        Creates and returns a selenium webelement for
        interfacing with the page.
        """
        if not accessors:
            accessors = self.accessors
        accessor = accessors

        by = accessor[0]
        value = accessor[1]
        el = self.page.driver.find_element(by=by, value=value)
        return el

    def click(self):
        return self.page.click(self.el)

    def click_and_go(self):
        time.sleep(random.uniform(3,6))
        return self.page.click_and_go(self.el)

    # def get_value(self):
    # def set_value(self, value):
    # def custom, e.g. zoom(self, x, y, dx, dy)

class TextField(PageObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        el = self.selenium_element(kwargs['accessors'])
        self.el = el

    def get_value(self):
        return self.page.get_el_value(self.el)

    def set_value(self, value):
        LOGGER.info('[%s]->set_text (%s)' % (self.name, value))
        time.sleep(random.uniform(3,5))
        return self.page.set_el_value(self.el, value)

class SelectField(PageObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        el = self.selenium_element(kwargs['accessors'])
        dropdown = selenium.webdriver.support.ui.Select(el)
        self.el = el
        self.dropdown = dropdown

    def get_value(self):
        return self.page.get_select_value(self.dropdown)

    def set_value(self, value):
        LOGGER.info('[%s]->set_select (%s)' % (self.name, value))
        time.sleep(random.uniform(6,11))
        return self.page.set_select_value(self.dropdown, value)

    def select_by_index(self, idx):
        return self.dropdown.select_by_index(idx)
        

class CheckField(PageObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        el = self.selenium_element(kwargs['accessors'])
        self.el = el

    def get_checked(self, ignore=False):
        return self.page.get_checkbox_value(self.el, ignore)

    def set_checked(self, checked):
        LOGGER.info('[%s]->set_checked (%s)' % (self.name, checked))
        time.sleep(random.uniform(2,3))
        return self.page.set_checkbox(self.el, checked)

class NullField(PageObject):
    def __init__(self, el, *args, **kwargs):
        self.el = el

    def __getattr__(self, attr):
        raise Exception

    def __bool__(self):
        return False

